module candump_top(
		mcp2515_0_spi_mosi, mcp2515_0_spi_miso, mcp2515_0_spi_clk, mcp2515_0_spi_cs_n,
		uart_tx,
		rst_n, clk,

		led_out, dipsw_in, button_in				// misc
	);

parameter	CLK_FREQ_MHZ	= 50;
parameter	CLK_FREQ_KHZ	= 50000;
parameter	CLK_FREQ_HZ		= 50000000;

// CAN (MCP2515)
output		mcp2515_0_spi_clk;		// IO13		CK_IO13		P17
input		mcp2515_0_spi_miso;		// IO12		CK_IO12		R17
output		mcp2515_0_spi_mosi;		// IO11		CK_IO11		U18
output		mcp2515_0_spi_cs_n;		// IO10		CK_IO10		V17

// output UART
output		uart_tx;				// PMOD-A-1(JA1)		G13

// clk, reset
input		rst_n;					// (active L) 
input		clk;					// 100MHz

// misc
output [3:0]	led_out;
input  [3:0]	dipsw_in;
input  [3:0]	button_in;

////////////////////////////////////////////////////////////////////////////////////
// global regs/wires
////////////////////////////////////////////////////////////////////////////////////
reg	   [7:0]	rst_n_sync_reg;
reg	   [7:0]	rst_n_cnt_reg;
wire			rst_n_sync;
wire			rst_p_sync;

// mcp2515_0 (CAN0)
wire [15:0]		can_0_id;					// note: 16bit
wire [7:0]		can_0_rtr, can_0_length;
wire [7:0]		can_0_data0, can_0_data1, can_0_data2, can_0_data3, can_0_data4, can_0_data5, can_0_data6, can_0_data7;
wire [31:0]		can_0_stat;
wire [31:0]		can_0_busdata_num, can_0_rcvdata_num;

wire [131:0]	can_0_din, can_0_dout;
wire			can_0_din_ena, can_0_din_req, can_0_din_ack;
wire			can_0_dout_ena, can_0_dout_req, can_0_dout_ack;
wire [15:0]		can_0_numdata, can_0_numspace;

// output uart
wire [9:0]		pct_dout;
wire [15:0]		pct_dout_numspace;
wire [15:0]		pct_dout_numpacket;
wire			pct_dout_wr;

// obc i/f
wire [3:0]		can_0_spd;

///////////////////////////////////////////////////////////////////////
// clock
///////////////////////////////////////////////////////////////////////
wire			sysclk;
reg				clk_reg;

always @(posedge clk) begin		// 100MHz
	clk_reg	<= ~clk_reg;
end

assign	sysclk		= clk_reg;	// 50MHz

///////////////////////////////////////////////////////////////////////
// reset
///////////////////////////////////////////////////////////////////////

// synchronizer
always @(posedge sysclk or negedge rst_n) begin
	if (rst_n == 1'b0) begin
		rst_n_sync_reg	<= 8'd0;
	end else begin
		rst_n_sync_reg	<= {rst_n_sync_reg[6:0], 1'b1};
	end
end

// timer
always @(posedge sysclk or negedge rst_n) begin
	if (rst_n == 1'b0) begin
		rst_n_cnt_reg	<= 8'd0;
	end else begin
		if (rst_n_sync_reg[7:2] == 6'b000000) begin
			rst_n_cnt_reg	<= 8'd0;
		end else if (rst_n_cnt_reg[7] == 1'b0) begin
			rst_n_cnt_reg	<= rst_n_cnt_reg + 1'b1;
		end
	end
end

assign	rst_n_sync	= rst_n_cnt_reg[7];
assign	rst_p_sync	= ~rst_n_cnt_reg[7];


///////////////////////////////////////////////////////////////////////
// 1us counter
///////////////////////////////////////////////////////////////////////
reg  [7:0]	tmr_1us_reg;

reg  [3:0]	cnt_1us_reg;
reg  [3:0]	cnt_10us_reg;
reg  [3:0]	cnt_100us_reg;
reg  [3:0]	cnt_1ms_reg;
reg  [3:0]	cnt_10ms_reg;
reg  [3:0]	cnt_100ms_reg;
reg  [3:0]	cnt_1s_reg;
reg  [3:0]	cnt_10s_reg;
reg  [3:0]	cnt_100s_reg;
reg  [3:0]	cnt_1000s_reg;
reg  [3:0]	cnt_10000s_reg;

wire [43:0]	time_info;
assign		time_info	= {
				cnt_10000s_reg,
				cnt_1000s_reg,
				cnt_100s_reg,
				cnt_10s_reg,
				cnt_1s_reg,
				cnt_100ms_reg,
				cnt_10ms_reg,
				cnt_1ms_reg,
				cnt_100us_reg,
				cnt_10us_reg,
				cnt_1us_reg
			};

always @(posedge sysclk) begin
	if (rst_n_sync == 1'b0) begin		// note; active L
		tmr_1us_reg		<= 8'd0;

		cnt_1us_reg		<= 4'd0;
		cnt_10us_reg	<= 4'd0;
		cnt_100us_reg	<= 4'd0;
		cnt_1ms_reg		<= 4'd0;
		cnt_10ms_reg	<= 4'd0;
		cnt_100ms_reg	<= 4'd0;
		cnt_1s_reg		<= 4'd0;
		cnt_10s_reg		<= 4'd0;
		cnt_100s_reg	<= 4'd0;
		cnt_1000s_reg	<= 4'd0;
		cnt_10000s_reg	<= 4'd0;

	end else begin
		if (tmr_1us_reg >= CLK_FREQ_MHZ - 1'b1) begin
			tmr_1us_reg	<= 8'd0;

			if (cnt_1us_reg < 4'd9) begin
				cnt_1us_reg	<= cnt_1us_reg + 1'b1;
			end else begin
				cnt_1us_reg	<= 4'd0;
				if (cnt_10us_reg < 4'd9) begin
					cnt_10us_reg	<= cnt_10us_reg + 1'b1;
				end else begin
					cnt_10us_reg	<= 4'd0;
					if (cnt_100us_reg < 4'd9) begin
						cnt_100us_reg	<= cnt_100us_reg + 1'b1;
					end else begin
						cnt_100us_reg	<= 4'd0;
						if (cnt_1ms_reg < 4'd9) begin
							cnt_1ms_reg	<= cnt_1ms_reg + 1'b1;
						end else begin
							cnt_1ms_reg	<= 4'd0;
							if (cnt_10ms_reg < 4'd9) begin
								cnt_10ms_reg	<= cnt_10ms_reg + 1'b1;
							end else begin
								cnt_10ms_reg	<= 4'd0;
								if (cnt_100ms_reg < 4'd9) begin
									cnt_100ms_reg	<= cnt_100ms_reg + 1'b1;
								end else begin
									cnt_100ms_reg	<= 4'd0;
									if (cnt_1s_reg < 4'd9) begin
										cnt_1s_reg	<= cnt_1s_reg + 1'b1;
									end else begin
										cnt_1s_reg	<= 4'd0;
										if (cnt_10s_reg < 4'd9) begin
											cnt_10s_reg	<= cnt_10s_reg + 1'b1;
										end else begin
											cnt_10s_reg	<= 4'd0;
											if (cnt_100s_reg < 4'd9) begin
												cnt_100s_reg	<= cnt_100s_reg + 1'b1;
											end else begin
												cnt_100s_reg	<= 4'd0;
												if (cnt_1000s_reg < 4'd9) begin
													cnt_1000s_reg	<= cnt_1000s_reg + 1'b1;
												end else begin
													cnt_1000s_reg	<= 4'd0;
													if (cnt_10000s_reg < 4'd9) begin
														cnt_10000s_reg	<= cnt_10000s_reg + 1'b1;
													end else begin
														cnt_10000s_reg	<= 4'd0;
													end	// 10000s
												end	// 1000s
											end	// 100s
										end	// 10s
									end	// 1s
								end	// 100ms
							end	// 10ms
						end	// 1ms
					end	// 100us
				end	// 10us
			end	// 1us

		end else begin
			tmr_1us_reg	<= tmr_1us_reg + 1'b1;
		end
	end
end


///////////////////////////////////////////////////////////////////////
// CAN #0 (SUB CAN, sensor side)
///////////////////////////////////////////////////////////////////////

//#define CANSPEED_125_16M 	7		// CAN speed at 125 kbps@16M osc	(CNF1 BRP, 16/Fosc)
//#define CANSPEED_250_16M	3		// CAN speed at 250 kbps@16M osc	(CNF1 BRP, 8/Fosc)
//#define CANSPEED_500_16M	1		// CAN speed at 500 kbps@16M osc	(CNF1 BRP, 4/Fosc)
//
//#define CANSPEED_125_8M 	3		// CAN speed at 125 kbps@8M osc		(CNF1 BRP, 8/Fosc)
//#define CANSPEED_250_8M	1		// CAN speed at 250 kbps@8M osc		(CNF1 BRP, 4/Fosc)
//#define CANSPEED_500_8M	0		// CAN speed at 500 kbps@8M osc		(CNF1 BRP, 2/Fosc)
assign	can_0_spd	= dipsw_in;


reg	[1:0]	mcp2515_0_spi_miso_sync_reg;
wire		mcp2515_0_spi_miso_in_sync;

always @(posedge sysclk) begin			// synchronizer
	if (rst_n_sync == 1'b0) begin		// note; active L
		mcp2515_0_spi_miso_sync_reg			<= 2'b11;
	end else begin
		mcp2515_0_spi_miso_sync_reg[1:0]	<= {mcp2515_0_spi_miso_sync_reg[0], mcp2515_0_spi_miso};
	end
end

wire		mcp2515_0_spi_access_req;
wire		mcp2515_0_spi_access_ack;
wire		mcp2515_0_spi_access_keep_cs_n;
wire [7:0]	mcp2515_0_spi_access_tx_data;
wire [7:0]	mcp2515_0_spi_access_rx_data;


spi_mcp2515_if u1 (
	.spi_mosi(mcp2515_0_spi_mosi),
	.spi_miso(mcp2515_0_spi_miso_sync_reg[1]),
	.spi_clk(mcp2515_0_spi_clk),
	.spi_cs_n(mcp2515_0_spi_cs_n),

	.tx_data8(mcp2515_0_spi_access_tx_data),
	.rx_data8(mcp2515_0_spi_access_rx_data),
	.req(mcp2515_0_spi_access_req),
	.ack(mcp2515_0_spi_access_ack),
	.keep_cs_n(mcp2515_0_spi_access_keep_cs_n),

	.clk(sysclk),
	.rst_n(rst_n_sync)
);


mcp2515_can_spi_mcp2515_can_spi u2 (
        .ap_clk(sysclk),
        .ap_rst(rst_p_sync),
        .ap_start(1'b1),
//      .ap_done,
//      .ap_idle,
//      .ap_ready,
        .dummy_tmr_out_i(1'b0),

        .spi_access_req(mcp2515_0_spi_access_req),
        .spi_access_ack(mcp2515_0_spi_access_ack),
        .spi_keep_cs_n(mcp2515_0_spi_access_keep_cs_n),
        .spi_tx_data(mcp2515_0_spi_access_tx_data),
        .spi_rx_data(mcp2515_0_spi_access_rx_data),

		.can_spd({4'b000, can_0_spd}),
        .can_id(can_0_id),			// note: 16bit
        .can_rtr(can_0_rtr),
        .can_length(can_0_length),
        .can_data0(can_0_data0),
        .can_data1(can_0_data1),
        .can_data2(can_0_data2),
        .can_data3(can_0_data3),
        .can_data4(can_0_data4),
        .can_data5(can_0_data5),
        .can_data6(can_0_data6),
        .can_data7(can_0_data7),

        .can_dout_ena(can_0_din_ena),
        .can_dout_req(can_0_din_req),
        .can_dout_ack(can_0_din_ack)
);

assign	can_0_din	= {
						time_info,		// bit[131:88]	time
						can_0_id,		// bit[87:72]	id
						can_0_length,	// bit[71:64]	len
						can_0_data0,	// bit[63:56]	d0
						can_0_data1,	// bit[55:48]	d1
						can_0_data2,	// bit[47:40]	d2
						can_0_data3,	// bit[39:32]	d3
						can_0_data4,	// bit[31:24]	d4
						can_0_data5,	// bit[23:16]	d5
						can_0_data6,	// bit[15:8]	d6
						can_0_data7		// bit[7:0]		d7
					};

fifo_mem #(
	.BIT_WIDTH(132),
	.FIFO_SIZE(4096)
) u3 (
	.din(can_0_din),
	.din_req(can_0_din_req),
	.din_ack(can_0_din_ack),
	.din_ena(can_0_din_ena),

	.dout(can_0_dout),
	.dout_req(can_0_dout_req),
	.dout_ack(can_0_dout_ack),
	.dout_ena(can_0_dout_ena),

	.num_data(can_0_numdata),
	.num_space(can_0_numspace),

	.clk(sysclk),
	.rst_n(rst_n_sync)
);


///////////////////////////////////////////////////////////////////////
// Packet constuctor
///////////////////////////////////////////////////////////////////////
data_formatter_data_formatter u4 (
        .ap_clk(sysclk),
        .ap_rst(rst_p_sync),
        .ap_start(1'b1),
//        .ap_done,
//        .ap_idle,
//        .ap_ready,

		// CAN0
        .can_0_sec(can_0_dout[131:112]),
        .can_0_secfrac(can_0_dout[111:88]),
        .can_0_id(can_0_dout[87:72]),
        .can_0_length(can_0_dout[71:64]),
        .can_0_data(can_0_dout[63:0]),

        .can_0_dout_ena(can_0_dout_ena),
        .can_0_dout_req(can_0_dout_req),
        .can_0_dout_ack(can_0_dout_ack),
        .can_0_numdata(can_0_numdata),

		// output
        .pct_dout(pct_dout),
        .pct_dout_numspace(pct_dout_numspace),
        .pct_dout_wr(pct_dout_wr)
);


///////////////////////////////////////////////////////////////////////
// Data sender
///////////////////////////////////////////////////////////////////////
uart_tx_packet #(
	.BUF_SIZE(32768)
) u5 (
	.uart_tx(uart_tx),
	.uart_cts_n(1'b0),		// fixed

//	.tx_spd(4'd9),			// 1Mbps
//	.tx_spd(4'd8),			// 2Mbps
	.tx_spd(4'd10),			// 3Mbps

	.tx_data(pct_dout),
	.tx_buf_wr(pct_dout_wr),
	.tx_buf_numspace(pct_dout_numspace),
	.tx_buf_numpacket(pct_dout_numpacket),

	.rst_sync_n(rst_n_sync),
	.clk50m(sysclk)
);


///////////////////////////////////////////////////////////////////////
// just for test
wire	ser_data_full;
assign	ser_data_full	= (pct_dout_numspace == 16'd0) ? 1'b1 : 1'b0;

wire	can_data_full;
assign	can_data_full	= (can_0_numspace == 16'd0) ? 1'b1 : 1'b0;

assign	led_out	= {can_0_din_req, can_data_full, ser_data_full, rst_n};

endmodule
