module fifo_mem(
			din, din_req, din_ack, din_ena,
			dout, dout_req, dout_ack, dout_ena,
			num_data, num_space,
			clk, rst_n
		);

parameter	BIT_WIDTH	= 96;
parameter	FIFO_SIZE	= 1024;				// MAX 2^(20 - 1)

// din side
input  [BIT_WIDTH - 1:0]	din;
input						din_req;
output						din_ack;
output						din_ena;

// dout side
output [BIT_WIDTH - 1:0]	dout;
input						dout_req;
output						dout_ack;
output						dout_ena;

output [15:0]				num_data;		// # of data (0 --- FIFO_SIZE)
output [15:0]				num_space;		// # of blank space (0 --- FIFO_SIZE)

input						clk;
input						rst_n;

////////////////////////////////////////////////////////////////////////
// main
////////////////////////////////////////////////////////////////////////
reg			din_ack_reg;
wire		din_ena_node;
reg  [19:0]	mem_wr_adr_reg;
wire		mem_wr_node;

reg			dout_ack_reg;
wire		dout_ena_node;
reg  [19:0]	mem_rd_adr_reg;
wire		mem_rd_node;

reg	 [20:0]	num_space_reg;

reg	 [BIT_WIDTH - 1:0]	mem[FIFO_SIZE - 1:0];	// memory body
reg	 [BIT_WIDTH - 1:0]	mem_out_reg;

// memmory body
always @(posedge clk) begin
	if (mem_wr_node)
		mem[mem_wr_adr_reg]	<= din;				// PI
end

always @(posedge clk) begin
	if (mem_rd_node)							// latch output
		mem_out_reg	<= mem[mem_rd_adr_reg];		// PO
end

// control
assign	din_ena_node	= (num_space_reg > 16'd32)		? 1'b1 : 1'b0;						// KEEP SPARE SPACE
assign	mem_wr_node		= (din_req == 1'b1 && din_ack_reg == 1'b0)		? 1'b1 : 1'b0;		// WRITE ANYWAY		NOTE: 1clock pulse width

assign	dout_ena_node	= (num_space_reg < FIFO_SIZE)	? 1'b1 : 1'b0;
assign	mem_rd_node		= (dout_ena_node == 1'b1 && dout_req == 1'b1 && dout_ack_reg == 1'b0)	? 1'b1 : 1'b0;		// NOTE: 1clock pulse width

always @(posedge clk) begin
	if (rst_n == 1'b0) begin
		num_space_reg	<= FIFO_SIZE;

		din_ack_reg		<= 1'b0;
		mem_wr_adr_reg	<= 20'd0;

		dout_ack_reg	<= 1'b0;
		mem_rd_adr_reg	<= 20'd0;

	end else begin
		// din
		if (din_ack_reg == 1'b1) begin
			if (din_req == 1'b0) begin
				din_ack_reg	<= 1'b0;
			end
		end else if (mem_wr_node == 1'b1) begin
			din_ack_reg		<= 1'b1;

			// inc addr ctr
			if (mem_wr_adr_reg < FIFO_SIZE - 1'b1) begin
				mem_wr_adr_reg	<= mem_wr_adr_reg + 1'b1;
			end else begin
				mem_wr_adr_reg	<= 20'd0;
			end
		end

		// dout
		if (dout_ack_reg == 1'b1) begin
			if (dout_req == 1'b0) begin
				dout_ack_reg	<= 1'b0;
			end
		end else if (mem_rd_node == 1'b1) begin
			dout_ack_reg	<= 1'b1;

			// inc addr ctr
			if (mem_rd_adr_reg < FIFO_SIZE - 1'b1) begin
				mem_rd_adr_reg	<= mem_rd_adr_reg + 1'b1;
			end else begin
				mem_rd_adr_reg	<= 20'd0;
			end
		end

		// count
		if (mem_wr_node == 1'b1 && mem_rd_node == 1'b0) begin
			if (num_space_reg > 21'd0) begin
				num_space_reg	<= num_space_reg - 1'b1;
			end
		end else if (mem_rd_node == 1'b1 && mem_wr_node == 1'b0) begin
			if (num_space_reg < FIFO_SIZE) begin
				num_space_reg	<= num_space_reg + 1'b1;
			end
		end
	end
end

////////////////////////////////////////////////////////////////////////
// PO
////////////////////////////////////////////////////////////////////////
assign	din_ack		= din_ack_reg;
assign	din_ena		= din_ena_node;

assign	dout		= mem_out_reg;
assign	dout_ack	= dout_ack_reg;
assign	dout_ena	= dout_ena_node;

assign	num_data	= FIFO_SIZE - num_space;
assign	num_space	= num_space_reg;

endmodule
