#include <ap_cint.h>
#include <ap_utils.h>
#include "synth_settings.h"
#include "mcp2515_defs.h"

typedef struct
{
	uint16	id;
	uint8	rtr;
	uint8 	length;
	uint8	data[8];
} tCAN;

////////////////////////////////////////////
// port
////////////////////////////////////////////

// SPI side
volatile uint1	spi_access_req	= 0;	// OUT
volatile uint1	spi_access_ack;			// IN
volatile uint1	spi_keep_cs_n	= 1;	// OUT
volatile uint8	spi_tx_data		= 0;	// OUT
volatile uint8	spi_rx_data;			// IN

// CAN side
volatile uint8	can_spd;				// IN

volatile uint16	can_id			= 0;
volatile uint8	can_rtr			= 0;
volatile uint8	can_length		= 0;
volatile uint8	can_data0		= 0;
volatile uint8	can_data1		= 0;
volatile uint8	can_data2		= 0;
volatile uint8	can_data3		= 0;
volatile uint8	can_data4		= 0;
volatile uint8	can_data5		= 0;
volatile uint8	can_data6		= 0;
volatile uint8	can_data7		= 0;

volatile uint1	can_dout_ena;			// IN
volatile uint1	can_dout_req	= 0;	// OUT
volatile uint1	can_dout_ack;			// IN

// dummy
volatile uint1	dummy_tmr_out	= 0;


////////////////////////////////////////////
// global
////////////////////////////////////////////
void wait_tmr(uint32 tmr)	// unit clk
{
	uint32	t;
	ap_wait();
	for (t = 0; t < tmr; t++) {
		dummy_tmr_out	= 1- dummy_tmr_out;
	}
	ap_wait();
}


////////////////////////////////////////////
// SPI I/F (important: SPECIFIC to MCP2515)
////////////////////////////////////////////

uint8 spi_putc(uint8 din, uint1 keep_cs_n)	// handshake with lower level verilog module
{
	uint8	retval;

	spi_tx_data		= din;
	spi_keep_cs_n	= keep_cs_n;
	do {
		ap_wait();
		spi_access_req	= 1;
	} while (spi_access_ack == 0);

	ap_wait();
	retval	= spi_rx_data;
	ap_wait();

	do {
		ap_wait();
		spi_access_req	= 0;
	} while (spi_access_ack == 1);

	return (retval);
}


void mcp2515_write_register(uint8 address, uint8 data)
{
	spi_putc(SPI_WRITE,	0);
	spi_putc(address,	0);
	spi_putc(data,		1);			// CS -> 1
}


uint8 mcp2515_read_register(uint8 address)
{
	uint8	ret;
	spi_putc(SPI_READ,	0);
	spi_putc(address,	0);
	ret	= spi_putc(0xFF, 1);		// CS -> 1
	return ret;
}


void mcp2515_bit_modify(uint8 address, uint8 mask, uint8 data)
{
	spi_putc(SPI_BIT_MODIFY,	0);
	spi_putc(address,			0);
	spi_putc(mask,				0);
	spi_putc(data,				1);	// CS -> 1
}


uint8 mcp2515_read_status(uint8 command)
{
	uint8	ret;
	spi_putc(command,	0);			// (spi command)
	ret	= spi_putc(0xFF, 1);		// CS -> 1
	return ret;
}


void mcp2515_init(uint8 speed)
{
	uint8	i;

	// reset MCP2515 by software reset.
	for (i = 0; i < 8; i++) {
		spi_putc(SPI_RESET,	1);			// CS -> 1
		wait_tmr(CLK_FREQ_MHZ * 100);	// 100us
	}

	// now in configuration mode
	mcp2515_bit_modify(CANINTF, (1 << RX0IF), 0);
	mcp2515_bit_modify(CANINTF, (1 << RX1IF), 0);

	// config CNF3,2,1,CANINTE,CANINTF (in this order, by multiple write)
	spi_putc(SPI_WRITE,	0);
	spi_putc(CNF3,		0);							// access start address
	spi_putc((1 << PHSEG21),	0);					// CNF3
	spi_putc((1 << BTLMODE) | (1 << PHSEG11),	0);	// CNF2
    spi_putc(speed,	0);								// CNF1
	spi_putc((1 << RX1IE) | (1 << RX0IE),	1);		// CANINTE, CS -> 1		enable int

	// deactivate the RXnBF Pins
	mcp2515_write_register(BFPCTRL, 0);

	// set TXnRTS as inputs
	mcp2515_write_register(TXRTSCTRL, 0);

	// turn off filters
	mcp2515_write_register(RXB0CTRL, (1 << RXM1) | (1 << RXM0));
	mcp2515_write_register(RXB1CTRL, (1 << RXM1) | (1 << RXM0));

	// to normal mode
	mcp2515_write_register(CANCTRL, 0);
}


uint8 mcp2515_get_message(tCAN *message, uint8 *st)
{
	uint8	intf;
	uint8	status;
	uint8	rx_cmd;
	uint16	d0, d1;
	uint8	length;
	uint8	t;
	uint8	buf_id;

	// read status (see manual figure 12-9)
	status	= mcp2515_read_status(SPI_RX_STATUS);
	*st		= status;

	if ((status & 0x40) != 0) {			// message in buffer 0
		rx_cmd	= SPI_READ_RX;			// buffer 0, rxb0sidh
		buf_id	= 0;
	}
	else if ((status & 0x80) != 0) {	// message in buffer 1
		rx_cmd	= SPI_READ_RX | 0x04;	// buffer 1, rxb0sidh
		buf_id	= 1;
	}
	else {
		return 0;
	}

	// read buffer
	spi_putc(rx_cmd,	0);

	// read id
	d0	= (uint16)spi_putc(0xFF, 0);
	d0	<<= 3;
	d1	= (uint16)spi_putc(0xFF, 0);
	d1	>>= 5;

	message->id	= d0 | d1;

	spi_putc(0xFF, 0);		// (discard)
	spi_putc(0xFF, 0);		// (discard)

	// read DLC
	length	= spi_putc(0xFF, 0);
	length	&= 0x0F;

	message->length	= length;
	message->rtr	= ((status & 0x08) != 0) ? 1 : 0;

	// read data
	for (t = 0; t < 8; t++) {
		message->data[t]	= 0;	// just clear buffer
	}
	if (length > 0 && length < 9) {
		for (t = 0; t < length - 1; t++) {
			message->data[t]	= spi_putc(0xFF, 0);
		}
		message->data[length - 1]	= spi_putc(0xFF, 1);	// CS -> 1
	}
	else {
		spi_putc(0xFF, 1);			// CS -> 1
	}

	// clear interrupt flag
	if (buf_id == 0) {
		mcp2515_bit_modify(CANINTF, (1 << RX0IF), 0);
	}
	else {
		mcp2515_bit_modify(CANINTF, (1 << RX1IF), 0);
	}

	return (status & 0x07) + 1;
}


////////////////////////////////////////////
// main
////////////////////////////////////////////
void mcp2515_can_spi()
{
#pragma HLS TOP
#pragma HLS INTERFACE ap_none port=dummy_tmr_out	register

#pragma HLS INTERFACE ap_none port=spi_access_req	register
#pragma HLS INTERFACE ap_none port=spi_access_ack
#pragma HLS INTERFACE ap_none port=spi_keep_cs_n	register
#pragma HLS INTERFACE ap_none port=spi_tx_data		register
#pragma HLS INTERFACE ap_none port=spi_rx_data

#pragma HLS INTERFACE ap_none port=can_spd
#pragma HLS INTERFACE ap_none port=can_id			register
#pragma HLS INTERFACE ap_none port=can_rtr			register
#pragma HLS INTERFACE ap_none port=can_length		register
#pragma HLS INTERFACE ap_none port=can_data0		register
#pragma HLS INTERFACE ap_none port=can_data1		register
#pragma HLS INTERFACE ap_none port=can_data2		register
#pragma HLS INTERFACE ap_none port=can_data3		register
#pragma HLS INTERFACE ap_none port=can_data4		register
#pragma HLS INTERFACE ap_none port=can_data5		register
#pragma HLS INTERFACE ap_none port=can_data6		register
#pragma HLS INTERFACE ap_none port=can_data7		register

#pragma HLS INTERFACE ap_none port=can_dout_ena
#pragma HLS INTERFACE ap_none port=can_dout_req		register
#pragma HLS INTERFACE ap_none port=can_dout_ack

	tCAN	message;
	char	exist_rcvdata;
	uint8	stat, err0, err1;
	uint32	busdata_num;
	uint32	rcvdata_num;

	////////////////////////////////////////
	// init ports
	////////////////////////////////////////
	ap_wait();
	spi_access_req	= 0;
	spi_keep_cs_n	= 1;
	can_dout_req	= 0;
	ap_wait();

	wait_tmr(CLK_FREQ_KHZ * 10);	// 10ms

	mcp2515_init(can_spd);

	exist_rcvdata	= 0;
	busdata_num		= 0;
	rcvdata_num		= 0;

	////////////////////////////////////////
	// main loop
	////////////////////////////////////////
	while (1) {
		// read mcp2515
		if (mcp2515_get_message(&message, &stat) != 0) {
			ap_wait();
			can_id		= message.id;
			can_rtr		= message.rtr;
			can_length	= message.length;
			can_data0	= message.data[0];
			can_data1	= message.data[1];
			can_data2	= message.data[2];
			can_data3	= message.data[3];
			can_data4	= message.data[4];
			can_data5	= message.data[5];
			can_data6	= message.data[6];
			can_data7	= message.data[7];

			exist_rcvdata	= 1;
			busdata_num++;
			ap_wait();
		}	// mcp2515

		// push data to fifo
		if (exist_rcvdata == 1 && can_dout_ena == 1) {
			// handshake
			do {
				can_dout_req	= 1;
				ap_wait();
			} while (can_dout_ack == 0);
			ap_wait();
			do {
				can_dout_req	= 0;
				ap_wait();
			} while (can_dout_ack == 1);
			ap_wait();

			exist_rcvdata	= 0;
			rcvdata_num++;
		}
	}
}
