module spi_mcp2515_if(
			spi_mosi, spi_miso, spi_clk, spi_cs_n,
			tx_data8, rx_data8, req, ack, keep_cs_n,
			clk, rst_n
		);

//	parameter	CLKNUM_100NS	= 16'd10;		// depend on clock speed (100MHz)
//	parameter	CLKNUM_100NS	= 16'd8;		// depend on clock speed (75MHz)
//	parameter	CLKNUM_100NS	= 16'd7;		// depend on clock speed (66MHz)
	parameter	CLKNUM_100NS	= 16'd5;		// depend on clock speed (50MHz)
//	parameter	CLKNUM_100NS	= 16'd2;		// depend on clock speed (20MHz)
//	parameter	CLKNUM_100NS	= 16'd1;		// depend on clock speed (10MHz)

//	parameter	CLKNUM_60NS	= 16'd6;			// depend on clock speed (100MHz)
//	parameter	CLKNUM_60NS	= 16'd5;			// depend on clock speed (75MHz)
//	parameter	CLKNUM_60NS	= 16'd4;			// depend on clock speed (66MHz)
	parameter	CLKNUM_60NS	= 16'd3;			// depend on clock speed (50MHz)

// ADIS side
output			spi_mosi;
input			spi_miso;			// NOTE; w/o synchronizer
output			spi_clk;
output			spi_cs_n;

// I/F side
input  [7:0]	tx_data8;
output [7:0]	rx_data8;
input			req;
output			ack;
input			keep_cs_n;			// keep cs (0) or not (1) after 1byte transfer

input			clk;
input			rst_n;

////////////////////////////////////////////////////////////////////////
// main
////////////////////////////////////////////////////////////////////////
reg			spi_mosi_reg, spi_clk_reg, spi_cs_n_reg;
reg [7:0]	tx_data_reg, rx_data_reg;
reg			ack_reg;

reg [4:0]	bitcnt_reg;		// 0-16
reg [15:0]	timer_reg;

parameter	ST_WAIT		= 3'b000;
parameter	ST_CS		= 3'b001;
parameter	ST_CLK0		= 3'b010;
parameter	ST_CLK1		= 3'b011;
parameter	ST_STALL	= 3'b100;
parameter	ST_ACK		= 3'b101;
reg [2:0]	state_reg;

always @(posedge clk) begin
	if (rst_n == 1'b0) begin
		spi_cs_n_reg	<= 1'b1;
		spi_clk_reg		<= 1'b0;
		spi_mosi_reg	<= 1'b0;
		ack_reg			<= 1'b0;
		state_reg		<= ST_WAIT;

	end else begin
		case (state_reg)
		/////////////////////////////////////////
		// wait request (clk=0, mosi=0)
		/////////////////////////////////////////
		ST_WAIT: begin
			spi_clk_reg		<= 1'b0;
			spi_mosi_reg	<= 1'b0;
			ack_reg			<= 1'b0;

			if (req == 1'b1) begin
				spi_cs_n_reg	<= 1'b0;

				tx_data_reg	<= tx_data8;		// load data
				rx_data_reg	<= 8'hFF;
				timer_reg	<= 16'd0;
				state_reg	<= ST_CS;
			end
		end

		/////////////////////////////////////////
		// assert cs
		/////////////////////////////////////////
		ST_CS: begin
			spi_clk_reg	<= 1'b0;

			// change output data at sclk falling edge
			spi_mosi_reg	<= tx_data_reg[7];
			tx_data_reg		<= {tx_data_reg[6:0], 1'b0};	// shift tx data

			timer_reg	<= 16'd0;
			bitcnt_reg	<= 5'd0;
			state_reg	<= ST_CLK0;
		end

		/////////////////////////////////////////
		// clk = L		0.06us
		/////////////////////////////////////////
		ST_CLK0: begin
			if (timer_reg >= CLKNUM_60NS - 1) begin
				spi_clk_reg	<= 1'b1;	// rising edge

				timer_reg	<= 16'd0;
				bitcnt_reg	<= bitcnt_reg + 1'b1;
				state_reg	<= ST_CLK1;
			end else begin
				timer_reg	<= timer_reg + 1'b1;
			end
		end

		/////////////////////////////////////////
		// clk = H		0.06us
		/////////////////////////////////////////
		ST_CLK1: begin
			if (timer_reg >= CLKNUM_60NS - 1) begin
				// load data
				rx_data_reg	<= {rx_data_reg[6:0], spi_miso};

				timer_reg	<= 16'd0;
				spi_clk_reg	<= 1'b0;	// falling edge

				if (bitcnt_reg >= 5'd8) begin
					spi_mosi_reg	<= 1'b0;

					state_reg		<= ST_STALL;
				end else begin
					// change output data at sclk falling edge
					spi_mosi_reg	<= tx_data_reg[7];
					tx_data_reg		<= {tx_data_reg[6:0], 1'b0};	// shift tx data

					state_reg		<= ST_CLK0;
				end
			end else begin
				timer_reg	<= timer_reg + 1'b1;
			end
		end

		/////////////////////////////////////////
		// deassert cs & stall	0.1us
		/////////////////////////////////////////
		ST_STALL: begin
			if (timer_reg >= CLKNUM_100NS - 1) begin
				spi_cs_n_reg	<= keep_cs_n;
				ack_reg			<= 1'b1;
				state_reg		<= ST_ACK;
			end else begin
				timer_reg	<= timer_reg + 1'b1;
			end
		end

		/////////////////////////////////////////
		// ack
		/////////////////////////////////////////
		ST_ACK: begin
			if (req == 1'b0) begin
				ack_reg		<= 1'b0;
				state_reg	<= ST_WAIT;
			end
		end

		default: begin
			spi_cs_n_reg	<= 1'b1;
			spi_clk_reg		<= 1'b0;
			spi_mosi_reg	<= 1'b0;
			ack_reg			<= 1'b0;
			state_reg		<= ST_WAIT;
		end
		endcase
	end
end

////////////////////////////////////////////////////////////////////////
// PO
////////////////////////////////////////////////////////////////////////
assign	spi_mosi	= spi_mosi_reg;
assign	spi_clk		= spi_clk_reg;
assign	spi_cs_n	= spi_cs_n_reg;

assign	rx_data8	= rx_data_reg;
assign	ack			= ack_reg;

endmodule
