module uart_tx_packet(
		uart_tx, uart_cts_n,
		tx_spd, tx_data, tx_buf_wr, tx_buf_numspace, tx_buf_numpacket,
		rst_sync_n, 
		clk50m
	);

parameter [15:0]	BUF_SIZE	= 16'd8192;		// MAX 65535 - 1; note bit width

output			uart_tx;				// UART rx output
input			uart_cts_n;

input  [3:0]	tx_spd;
input  [9:0]	tx_data;
input			tx_buf_wr;				// 1 pulse
output [15:0]	tx_buf_numspace;
output [15:0]	tx_buf_numpacket;

input			rst_sync_n;				// sync
input			clk50m;					// fixed freq

/////////////////////////////////////////////////
// functions/definitions
/////////////////////////////////////////////////
wire			clk;
assign			clk	= clk50m;

function [19:0] spd2clk50;		// 50MHz ... 20ns
	input	[3:0]	spd;
	reg		[19:0]	ret;
begin
	case (spd)
	// width of one pulse (unit: clk)
	4'd0:		ret	= 20'd41667;		// 1200			833333ns
	4'd1:		ret	= 20'd5208;			// 9600			104166ns
	4'd2:		ret	= 20'd3472;			// 14400		69444ns
	4'd3:		ret	= 20'd1302;			// 38400		26042ns
	4'd4:		ret	= 20'd868;			// 57600		17361ns
	4'd5:		ret	= 20'd434;			// 115200		8681ns
	4'd6:		ret	= 20'd217;			// 230400		4340ns
	4'd7:		ret	= 20'd109;			// 460800		2170ns
	4'd8:		ret	= 20'd50;			// 1000000		1000ns
	4'd9:		ret	= 20'd25;			// 2000000		500ns
	4'd10:		ret	= 20'd17;			// 3000000		333ns
	4'd11:		ret	= 20'd13;			// 4000000		250ns
	4'd12:		ret	= 20'd10;			// 5000000		200ns
	4'd13:		ret	= 20'd6;			// 8000000		125ns
	default:	ret	= 20'd0;
	endcase

	spd2clk50	= ret;
end
endfunction

wire [19:0]		spd_node;
wire [19:0]		spd_half_node;
assign			spd_node		= spd2clk50(tx_spd);

reg	 [1:0]		uart_tx_reg;
reg	 [1:0]		uart_cts_n_sync_reg;
wire			uart_cts_n_sync_node;

reg	 [19:0]		tx_tmr_reg;
reg	 [3:0]		tx_bitcnt_reg;
reg	 [7:0]		tx_data_reg;

parameter [2:0]	ST_WAIT_DATA	= 3'd0;
parameter [2:0]	ST_RD_BUFDATA	= 3'd1;
parameter [2:0]	ST_SEND_START	= 3'd2;
parameter [2:0]	ST_SEND_DATA	= 3'd3;
parameter [2:0]	ST_SEND_STOP	= 3'd4;
reg	 [2:0]		state_reg;

reg	 [15:0]		buf_dnum_reg;				// note: bit width (0-255)
reg	 [15:0]		buf_numpacket_reg;
reg	 [9:0]		buf_mem[BUF_SIZE - 1:0];
reg	 [15:0]		buf_wadr_reg;
reg	 [15:0]		buf_radr_reg;
wire [9:0]		buf_rd_data;

wire			buf_we, buf_rd, buf_empty, buf_full;
assign			buf_empty	= (buf_dnum_reg == 16'd0)							? 1'b1 : 1'b0;
assign			buf_full	= (buf_dnum_reg >= BUF_SIZE)						? 1'b1 : 1'b0;
assign			buf_we		= (tx_buf_wr == 1'b1 && buf_full == 1'b0)			? 1'b1 : 1'b0;
assign			buf_rd		= (state_reg == ST_RD_BUFDATA && buf_empty == 1'b0)	? 1'b1 : 1'b0;
assign			buf_rd_data	= buf_mem[buf_radr_reg];

/////////////////////////////////////////////////
// sender
/////////////////////////////////////////////////

// sync
always @(posedge clk) begin
	uart_cts_n_sync_reg	<= {uart_cts_n_sync_reg[0], uart_cts_n};
end

assign	uart_cts_n_sync_node	= uart_cts_n_sync_reg[1];

// ctrl
always @(posedge clk) begin
	if (rst_sync_n == 1'b0) begin
		uart_tx_reg	<= 1'b1;
		state_reg	<= ST_WAIT_DATA;

	end else begin
		case (state_reg)
		/////////////////////////////////////
		// wait start bit
		/////////////////////////////////////
		ST_WAIT_DATA: begin
			uart_tx_reg	<= 1'b1;
			if (buf_numpacket_reg > 16'd0 && buf_empty != 1'b1 && uart_cts_n_sync_node == 1'b0) begin
				state_reg	<= ST_RD_BUFDATA;
			end
		end

		/////////////////////////////////////
		// read data from buffer
		/////////////////////////////////////
		ST_RD_BUFDATA: begin
			tx_data_reg		<= buf_rd_data[7:0];

			uart_tx_reg		<= 1'b0;				// start bit
			tx_tmr_reg		<= 20'd2;
			state_reg		<= ST_SEND_START;
		end

		/////////////////////////////////////
		// send start bit
		/////////////////////////////////////
		ST_SEND_START: begin
			if (tx_tmr_reg >= spd_node) begin	// wait 1 pulse duration
				uart_tx_reg		<= tx_data_reg[0];
				tx_data_reg		<= {1'b1, tx_data_reg[7:1]};		// RIGHT SHIFT (LSB 1ST)
				tx_bitcnt_reg	<= 4'd0;
				tx_tmr_reg		<= 20'd0;

				state_reg		<= ST_SEND_DATA;
			end else begin
				tx_tmr_reg		<= tx_tmr_reg + 1'b1;
			end
		end

		/////////////////////////////////////
		// send data
		/////////////////////////////////////
		ST_SEND_DATA: begin
			if (tx_tmr_reg >= spd_node) begin	// wait 1 pulse duration
				uart_tx_reg		<= tx_data_reg[0];
				tx_data_reg		<= {1'b1, tx_data_reg[7:1]};		// RIGHT SHIFT (LSB 1ST)
				tx_bitcnt_reg	<= tx_bitcnt_reg + 1'b1;
				tx_tmr_reg		<= 20'd1;

				if (tx_bitcnt_reg >= 4'd7) begin
					state_reg		<= ST_SEND_STOP;
				end
			end else begin
				tx_tmr_reg		<= tx_tmr_reg + 1'b1;
			end
		end

		/////////////////////////////////////
		// send stop bit
		/////////////////////////////////////
		ST_SEND_STOP: begin
			if (tx_tmr_reg >= spd_node) begin	// wait 1 pulse duration
				uart_tx_reg		<= 1'b1;
				state_reg		<= ST_WAIT_DATA;
			end else begin
				tx_tmr_reg		<= tx_tmr_reg + 1'b1;
			end
		end

		default: begin
			uart_tx_reg	<= 1'b1;
			state_reg	<= ST_WAIT_DATA;
		end
		endcase
	end
end


/////////////////////////////////////////////////
// buffer
/////////////////////////////////////////////////

// buffer body
always @(posedge clk) begin
	if (buf_we) begin
		buf_mem[buf_wadr_reg]	<= tx_data;
	end
end

// buffer access control
always @(posedge clk) begin
	if (rst_sync_n == 1'b0) begin
		buf_dnum_reg	<= 16'd0;
		buf_wadr_reg	<= 16'd0;
		buf_radr_reg	<= 16'd0;

	end else begin
		// write request
		if (buf_we == 1'b1) begin
			if (buf_wadr_reg < BUF_SIZE - 1) begin
				buf_wadr_reg	<= buf_wadr_reg + 1'b1;
			end else begin
				buf_wadr_reg	<= 16'd0;
			end
		end

		// read request
		if (buf_rd == 1'b1) begin
			if (buf_radr_reg < BUF_SIZE - 1) begin
				buf_radr_reg	<= buf_radr_reg + 1'b1;
			end else begin
				buf_radr_reg	<= 16'd0;
			end
		end

		// # of existing words
		if (buf_we == 1'b1 && buf_rd == 1'b0) begin
			buf_dnum_reg		<= buf_dnum_reg + 1'b1;
		end else if (buf_rd == 1'b1 && buf_we == 1'b0) begin
			buf_dnum_reg		<= buf_dnum_reg - 1'b1;
		end
	end
end

// count # of packet
wire	end_of_wr_packet	= (buf_we == 1'b1 && tx_data[8] == 1'b1)					? 1'b1 : 1'b0;
wire	end_of_rd_packet	= (state_reg == ST_RD_BUFDATA && buf_rd_data[8] == 1'b1)	? 1'b1 : 1'b0;

always @(posedge clk) begin
	if (rst_sync_n == 1'b0) begin
		buf_numpacket_reg	<= 16'd0;

	end else begin
		if (end_of_wr_packet == 1'b1 && end_of_rd_packet == 1'b0) begin
			buf_numpacket_reg	<= buf_numpacket_reg + 1'b1;
		end else if (end_of_rd_packet == 1'b1 && end_of_wr_packet == 1'b0) begin
			buf_numpacket_reg	<= buf_numpacket_reg - 1'b1;
		end
	end
end

/////////////////////////////////////////////////
// PO
/////////////////////////////////////////////////
assign	uart_tx		= uart_tx_reg;

assign	tx_buf_numspace		= (buf_full == 1'b1)	? 16'd0 : (BUF_SIZE - buf_dnum_reg);	// for safety
assign	tx_buf_numpacket	= buf_numpacket_reg;

endmodule
